@@ -1,6 +1,6 @@
 use crate::svd::{
-    Access, BitRange, EnumeratedValues, Field, Peripheral, Register, RegisterProperties, Usage,
-    WriteConstraint,
+    Access, BitRange, DeriveFrom, EnumeratedValues, Field, Peripheral, Register,
+    RegisterProperties, Usage, WriteConstraint,
 };
 use cast::u64;
 use core::u64;
@@ -19,15 +19,15 @@ pub fn render(
     defs: &RegisterProperties,
     config: &Config,
 ) -> Result<TokenStream> {
-    let access = util::access_of(register);
+    let properties = register.properties.derive_from(defs);
+    let access = util::access_of(&properties, register.fields.as_deref());
     let name = util::name_of(register, config.ignore_groups);
     let span = Span::call_site();
     let name_pc = Ident::new(&name.to_sanitized_upper_case(), span);
     let name_uc_spec = Ident::new(&format!("{}_SPEC", &name.to_sanitized_upper_case()), span);
     let name_sc = Ident::new(&name.to_sanitized_snake_case(), span);
-    let rsize = register
+    let rsize = properties
         .size
-        .or(defs.size)
         .ok_or_else(|| anyhow!("Register {} has no `size` field", register.name))?;
     let rsize = if rsize < 8 {
         8
@@ -44,16 +44,15 @@ pub fn render(
         }))
         .as_ref(),
     );
-    let res_val = register.reset_value.or(defs.reset_value).map(|v| v as u64);
 
     let mut mod_items = TokenStream::new();
     let mut r_impl_items = TokenStream::new();
     let mut w_impl_items = TokenStream::new();
     let mut methods = vec![];
 
-    let can_read = [Access::ReadOnly, Access::ReadWriteOnce, Access::ReadWrite].contains(&access);
-    let can_write = access != Access::ReadOnly;
-    let can_reset = res_val.is_some();
+    let can_read = access.can_read();
+    let can_write = access.can_write();
+    let can_reset = properties.reset_value.is_some();
 
     if can_read {
         let desc = format!("Register `{}` reader", register.name);
@@ -137,8 +136,8 @@ pub fn render(
                 peripheral,
                 all_peripherals,
                 &rty,
-                res_val,
                 access,
+                &properties,
                 &mut mod_items,
                 &mut r_impl_items,
                 &mut w_impl_items,
@@ -163,14 +162,40 @@ pub fn render(
 
         mod_items.extend(w_impl_items);
 
-        mod_items.extend(quote! {
-            #[doc = "Writes raw bits to the register."]
-            #[inline(always)]
-            pub unsafe fn bits(&mut self, bits: #rty) -> &mut Self {
-                self.0.bits(bits);
-                self
+        // the writer can be safe if:
+        // * there is a single field that covers the entire register
+        // * that field can represent all values
+        let can_write_safe = match register
+            .fields
+            .as_ref()
+            .and_then(|fields| fields.iter().next())
+            .and_then(|field| field.write_constraint)
+        {
+            Some(WriteConstraint::Range(range)) => {
+                range.min == 0 && range.max == u64::MAX >> (64 - rsize)
             }
-        });
+            _ => false,
+        };
+
+        if can_write_safe {
+            mod_items.extend(quote! {
+                #[doc = "Writes raw bits to the register."]
+                #[inline(always)]
+                pub fn bits(&mut self, bits: #rty) -> &mut Self {
+                    unsafe { self.0.bits(bits) };
+                    self
+                }
+            });
+        } else {
+            mod_items.extend(quote! {
+                #[doc = "Writes raw bits to the register."]
+                #[inline(always)]
+                pub unsafe fn bits(&mut self, bits: #rty) -> &mut Self {
+                    self.0.bits(bits);
+                    self
+                }
+            });
+        }
 
         close.to_tokens(&mut mod_items);
     }
@@ -231,7 +256,7 @@ pub fn render(
             }
         });
     }
-    if let Some(rv) = res_val.map(util::hex) {
+    if let Some(rv) = properties.reset_value.map(util::hex) {
         let doc = format!("`reset()` method sets {} to value {}", register.name, &rv);
         mod_items.extend(quote! {
             #[doc = #doc]
@@ -264,16 +289,16 @@ pub fn fields(
     peripheral: &Peripheral,
     all_peripherals: &[Peripheral],
     rty: &Ident,
-    reset_value: Option<u64>,
     access: Access,
+    properties: &RegisterProperties,
     mod_items: &mut TokenStream,
     r_impl_items: &mut TokenStream,
     w_impl_items: &mut TokenStream,
     config: &Config,
 ) -> Result<()> {
     let span = Span::call_site();
-    let can_read = [Access::ReadOnly, Access::ReadWriteOnce, Access::ReadWrite].contains(&access);
-    let can_write = access != Access::ReadOnly;
+    let can_read = access.can_read();
+    let can_write = access.can_write();
 
     // TODO enumeratedValues
     let inline = quote! { #[inline(always)] };
@@ -284,11 +309,8 @@ pub fn fields(
         let name_sc = Ident::new(&name.to_sanitized_snake_case(), span);
         let name_pc = name.to_sanitized_upper_case();
         let bits = Ident::new(if width == 1 { "bit" } else { "bits" }, span);
-        let description = if let Some(d) = &f.description {
-            util::respace(&util::escape_brackets(d))
-        } else {
-            "".to_owned()
-        };
+        let description_raw = f.description.as_deref().unwrap_or(""); // raw description, if absent using empty string
+        let description = util::respace(&util::escape_brackets(description_raw));
 
         let can_read = can_read
             && (f.access != Some(Access::WriteOnly))
@@ -296,12 +318,18 @@ pub fn fields(
         let can_write = can_write && (f.access != Some(Access::ReadOnly));
 
         let mask = u64::MAX >> (64 - width);
-        let hexmask = &util::hex(mask);
+        let hexmask = &util::digit_or_hex(mask);
         let offset = u64::from(offset);
-        let rv = reset_value.map(|rv| (rv >> offset) & mask);
+        let rv = properties.reset_value.map(|rv| (rv >> offset) & mask);
         let fty = width.to_ty()?;
         let evs = &f.enumerated_values;
 
+        let use_mask = if let Some(size) = properties.size {
+            size != width
+        } else {
+            true
+        };
+
         let lookup_results = lookup(
             evs,
             fields,
@@ -320,7 +348,7 @@ pub fn fields(
 
         let field_dim = match f {
             Field::Array(_, de) => {
-                let (first, index) = if let Some(dim_index) = &de.dim_index {
+                let first = if let Some(dim_index) = &de.dim_index {
                     if let Ok(first) = dim_index[0].parse::<u32>() {
                         let sequential_indexes = dim_index
                             .iter()
@@ -329,17 +357,14 @@ pub fn fields(
                         if !sequential_indexes {
                             return Err(anyhow!("unsupported array indexes in {}", f.name));
                         }
-                        (first, None)
+                        first
                     } else {
-                        (0, de.dim_index.clone())
+                        0
                     }
                 } else {
-                    (0, None)
-                };
-                let suffixes: Vec<_> = match index {
-                    Some(ix) => ix,
-                    None => (0..de.dim).map(|i| (first + i).to_string()).collect(),
+                    0
                 };
+                let suffixes: Vec<_> = de.indexes().collect();
                 let suffixes_str = format!("({}-{})", first, first + de.dim - 1);
                 Some((first, de.dim, de.dim_increment, suffixes, suffixes_str))
             }
@@ -374,14 +399,18 @@ pub fn fields(
                 quote! {
                     ((self.bits >> #offset) & #hexmask) #cast
                 }
-            } else {
+            } else if use_mask {
                 quote! {
                     (self.bits & #hexmask) #cast
                 }
+            } else {
+                quote! {
+                    self.bits
+                }
             };
 
             if let Some((first, dim, increment, suffixes, suffixes_str)) = &field_dim {
-                let offset_calc = calculate_offset(*first, *increment, offset);
+                let offset_calc = calculate_offset(*first, *increment, offset, true);
                 let value = quote! { ((self.bits >> #offset_calc) & #hexmask) #cast };
                 let doc = &util::replace_suffix(&description, suffixes_str);
                 r_impl_items.extend(quote! {
@@ -398,18 +427,22 @@ pub fn fields(
                         quote! {
                             ((self.bits >> #sub_offset) & #hexmask) #cast
                         }
-                    } else {
+                    } else if use_mask {
                         quote! {
                             (self.bits & #hexmask) #cast
                         }
+                    } else {
+                        quote! {
+                            self.bits
+                        }
                     };
                     let name_sc_n = Ident::new(
-                        &util::replace_suffix(&f.name.to_sanitized_snake_case(), &suffix),
+                        &util::replace_suffix(&f.name, suffix).to_sanitized_snake_case(),
                         Span::call_site(),
                     );
                     let doc = util::replace_suffix(
-                        &description_with_bits(&description, sub_offset, width),
-                        &suffix,
+                        &description_with_bits(description_raw, sub_offset, width),
+                        suffix,
                     );
                     r_impl_items.extend(quote! {
                         #[doc = #doc]
@@ -420,7 +453,7 @@ pub fn fields(
                     });
                 }
             } else {
-                let doc = description_with_bits(&description, offset, width);
+                let doc = description_with_bits(description_raw, offset, width);
                 r_impl_items.extend(quote! {
                     #[doc = #doc]
                     #inline
@@ -527,6 +560,7 @@ pub fn fields(
                         pub struct #name_pc_r(crate::FieldReader<#fty, #name_pc_a>);
 
                         impl #name_pc_r {
+                            #[inline(always)]
                             pub(crate) fn new(bits: #fty) -> Self {
                                 #name_pc_r(crate::FieldReader::new(bits))
                             }
@@ -549,6 +583,7 @@ pub fn fields(
                     pub struct #name_pc_r(crate::FieldReader<#fty, #fty>);
 
                     impl #name_pc_r {
+                        #[inline(always)]
                         pub(crate) fn new(bits: #fty) -> Self {
                             #name_pc_r(crate::FieldReader::new(bits))
                         }
@@ -587,7 +622,7 @@ pub fn fields(
                         let pc = util::replace_suffix(base.field, "");
                         let pc = pc.to_sanitized_upper_case();
                         let base_pc_w = Ident::new(&(pc + "_AW"), span);
-                        derive_from_base(mod_items, &base, &name_pc_aw, &base_pc_w, &description)
+                        derive_from_base(mod_items, &base, name_pc_aw, &base_pc_w, &description)
                     } else if variants.is_empty() {
                         add_with_no_variants(mod_items, name_pc_aw, &fty, &description, rv);
                     } else {
@@ -679,11 +714,17 @@ pub fn fields(
                         }
                     }
                 } else {
+                    let op = if use_mask {
+                        quote!{ self.w.bits = (self.w.bits & !#hexmask) | (value as #rty & #hexmask); }
+                    } else {
+                        quote!{ self.w.bits = value; }
+                    };
+
                     quote! {
                         ///Writes raw bits to the field
                         #inline
                         pub #unsafety fn #bits(self, value: #fty) -> &'a mut W {
-                            self.w.bits = (self.w.bits & !#hexmask) | (value as #rty & #hexmask);
+                            #op
                             self.w
                         }
                     }
@@ -749,7 +790,7 @@ pub fn fields(
             }
 
             if let Some((first, dim, increment, suffixes, suffixes_str)) = &field_dim {
-                let offset_calc = calculate_offset(*first, *increment, offset);
+                let offset_calc = calculate_offset(*first, *increment, offset, false);
                 let doc = &util::replace_suffix(&description, suffixes_str);
                 w_impl_items.extend(quote! {
                     #[doc = #doc]
@@ -761,12 +802,12 @@ pub fn fields(
                 for (i, suffix) in (0..*dim).zip(suffixes.iter()) {
                     let sub_offset = offset + (i as u64) * (*increment as u64);
                     let name_sc_n = Ident::new(
-                        &util::replace_suffix(&f.name.to_sanitized_snake_case(), &suffix),
+                        &util::replace_suffix(&f.name, suffix).to_sanitized_snake_case(),
                         Span::call_site(),
                     );
                     let doc = util::replace_suffix(
-                        &description_with_bits(&description, sub_offset, width),
-                        &suffix,
+                        &description_with_bits(description_raw, sub_offset, width),
+                        suffix,
                     );
                     let sub_offset = util::unsuffixed(sub_offset as u64);
                     if !config.const_generic {
@@ -788,7 +829,7 @@ pub fn fields(
                     }
                 }
             } else {
-                let doc = description_with_bits(&description, offset, width);
+                let doc = description_with_bits(description_raw, offset, width);
                 w_impl_items.extend(quote! {
                     #[doc = #doc]
                     #inline
@@ -908,7 +949,7 @@ fn add_from_variants(
 
     let mut vars = TokenStream::new();
     for v in variants.iter().map(|v| {
-        let desc = util::escape_brackets(&format!("{}: {}", v.value, v.doc));
+        let desc = util::escape_brackets(&util::respace(&format!("{}: {}", v.value, v.doc)));
         let pcv = &v.pc;
         let pcval = &util::unsuffixed(v.value);
         quote! {
@@ -941,7 +982,12 @@ fn add_from_variants(
     });
 }
 
-fn calculate_offset(first: u32, increment: u32, offset: u64) -> TokenStream {
+fn calculate_offset(
+    first: u32,
+    increment: u32,
+    offset: u64,
+    with_parentheses: bool,
+) -> TokenStream {
     let mut res = if first != 0 {
         let first = util::unsuffixed(first as u64);
         quote! { n - #first }
@@ -960,7 +1006,12 @@ fn calculate_offset(first: u32, increment: u32, offset: u64) -> TokenStream {
         let offset = &util::unsuffixed(offset);
         res = quote! { #res + #offset };
     }
-    res
+    let single_ident = (first == 0) && (increment == 1) && (offset == 0);
+    if with_parentheses && !single_ident {
+        quote! { (#res) }
+    } else {
+        res
+    }
 }
 
 fn description_with_bits(description: &str, offset: u64, width: u32) -> String {
@@ -1213,7 +1264,7 @@ fn lookup_in_peripherals<'p>(
     all_peripherals: &'p [Peripheral],
 ) -> Result<(&'p EnumeratedValues, Option<Base<'p>>)> {
     if let Some(peripheral) = all_peripherals.iter().find(|p| p.name == base_peripheral) {
-        let all_registers = peripheral.reg_iter().collect::<Vec<_>>();
+        let all_registers = peripheral.all_registers().collect::<Vec<_>>();
         lookup_in_peripheral(
             Some(base_peripheral),
             base_register,
